#!/usr/bin/env python
# coding: utf8

"""Visual keyboard highly customizable for pygame."""

from .vkeyboard import VKeyboardLayout, VKeyboard
from .vrenderers import VKeyboardRenderer

__version__ = '2.0.9'



#!/usr/bin/env python
# coding: utf8

"""
Renderer for the keyboard and the text box used to display
the current text.
"""

import os.path as osp
import pygame  # pylint: disable=import-error

from . import vkeys


def fit_font(font_name, max_height):
    """Set the size of the font to fit the given height.

    This function uses the binary search algorithm to go faster
    than a one-by-one try.

    Parameters
    ----------
    font_name:
        Path to font file for rendering key.
    max_height:
        Height to fit.
    """
    font = pygame.font.Font(font_name, 1)

    # Ensure a large panel of characters heights
    text = "?/|!()§&@0123456789azertyuiopqsdfghjklmwxcvbnAZERTYUIOPQSDFGHJKLMWXCVBN"  # noqa

    start = 1
    end = max_height * 2

    while start < end:
        k = (start + end) // 2
        font = pygame.font.Font(font_name, k)
        height = font.size(text)[1]
        if height > max_height:
            end = k
        else:
            start = k + 1
        if start < end:
            # Run garbage collector, to avoid opening too many files
            del font

    return font


def draw_round_rect(surface, color, rect, radius=0.1, width=0):
    """Draw a rounded rectangle.

    Parameters
    ----------
    surface:
        Surface to draw on.
    color:
        RGBA tuple color to draw with, the alpha value is optional.
    rect:
        Rectangle to draw, position and dimensions.
    radius:
        Used for drawing rectangle with rounded corners. The supported range is
        [0, 1] with 0 representing a rectangle without rounded corners.
    width:
        Line thickness (0 to fill the rectangle).
    """
    rect = pygame.Rect(rect)
    if len(color) == 4:
        alpha = color[-1]
        color = color[:3] + (0,)
    else:
        alpha = 255
        color += (0,)

    shape = pygame.Surface(rect.size, pygame.SRCALPHA)

    circle = pygame.Surface([min(rect.size) * 3] * 2, pygame.SRCALPHA)
    if width > 0:
        pygame.draw.arc(circle, (0, 0, 0), circle.get_rect(),
                        1.571, 3.1415, width * 8)
    else:
        pygame.draw.ellipse(circle, (0, 0, 0), circle.get_rect(), 0)
    circle = pygame.transform.smoothscale(circle,
                                          [int(min(rect.size) * radius)] * 2)

    i = 1
    shape_rect = shape.get_rect()
    for pos in ('topleft', 'topright', 'bottomleft', 'bottomright'):
        r = circle.get_rect(**{pos: getattr(shape_rect, pos)})
        shape.blit(circle, r)
        if width > 0:
            circle = pygame.transform.rotate(circle, -i * 90)
        i += 1

    hrect = shape_rect.inflate(0, -circle.get_height() + 1)
    vrect = shape_rect.inflate(-circle.get_width() + 1, 0)
    if width > 0:
        hrect.width = width
        vrect.height = width
        shape.fill((0, 0, 0), hrect)
        shape.fill((0, 0, 0), vrect)
        hrect.right = shape_rect.right
        vrect.bottom = shape_rect.bottom
        shape.fill((0, 0, 0), hrect)
        shape.fill((0, 0, 0), vrect)
    else:
        shape.fill((0, 0, 0), hrect)
        shape.fill((0, 0, 0), vrect)

    shape.fill(color, special_flags=pygame.BLEND_RGBA_MAX)
    shape.fill((255, 255, 255, alpha), special_flags=pygame.BLEND_RGBA_MIN)

    return surface.blit(shape, rect)


class VKeyboardRenderer(object):
    """
    A VKeyboardRenderer is in charge of keyboard rendering.

    It handles keyboard rendering properties such as color or padding,
    and provides several rendering methods.

    .. note::
        A DEFAULT and DARK styles are available as class attribute.
    """

    DEFAULT = None
    DARK = None

    def __init__(self,
                 font_name,
                 text_color,
                 cursor_color,
                 selection_color,
                 background_color,
                 background_key_color,
                 background_input_color,
                 text_special_key_color=None,
                 background_special_key_color=None):
        """VKeyboardStyle default constructor.

        Some parameters take a list of color tuples, one per state.
        The states are: (released, pressed)

        Parameters
        ----------
        font_name:
            Path to font file for rendering key.
        text_color:
            List of RGB tuples for text color (one tuple per state).
        cursor_color:
            RGB tuple for cursor color of text input.
        selection_color:
            RGB tuple for selected key color.
        background_color:
            RGB tuple for background colo   r for text.
        background_key_color:
            List of RGB tuples for key background color (one tuple per state).
        background_input_color:
            RGB tuple for background color of the text input.
        text_special_key_color:
            List of RGB tuples for special key text color (one tuple per state).
        background_special_key_color:
            List of RGB tuples for special key background color (one tuple per
            state).
        """
        self.font = None
        self.font_height = None
        self.font_input = None
        self.font_input_height = None
        self.font_name = font_name
        self.text_color = text_color
        self.cursor_color = cursor_color
        self.selection_color = selection_color
        self.background_color = background_color
        self.background_key_color = background_key_color
        self.background_input_color = background_input_color

        self.text_special_key_color = text_special_key_color
        self.background_special_key_color = background_special_key_color

    def get_text_width(self, text):
        """Return the width of the given text in the text input box.

        Parameters
        ----------
        text:
            Text to evaluate.
        """
        return self.font_input.size(text)[0]

    def truncate(self, text, max_width, start=0, nearest=False):
        """Truncate the given text in order to fit the maximum
        given width.

        This function uses the binary search algorithm to go faster
        than a one-by-one try.

        Parameters
        ----------
        text:
            Text to split.
        max_width:
            Maximum authorized width of the text (according to font).
        start:
            Index for searching the text part with correct width.
        nearest:
            If True, the returned text can have a width higher than
            the ``max_width`` to reduce abs(max_width - width).

        Returns
        -------
        (part, width):
            Truncated text and its rendered width.
        """
        width = 0
        end = len(text)

        if end < start:
            return text, self.get_text_width(text)

        while start < end:
            k = (start + end) // 2
            new_width = self.get_text_width(text[:k+1])
            if new_width > max_width:
                end = k
            else:
                width = new_width
                start = k + 1

        if nearest:
            next_width = self.get_text_width(text[:start+1])
            if abs(max_width - next_width) < abs(max_width - width):
                return text[:start+1], next_width

        return text[:start], width

    def draw_background(self, surface):
        """Default drawing method for background.

        Background is drawn as a simple rectangle filled using this
        style background color attribute.

        Parameters
        ----------
        surface:
            Surface background should be drawn in.
        """
        surface.fill(self.background_color)

    def draw_cursor(self, surface, cursor):
        """Default drawing method for cursor of the text input box.

        Cursor is drawn as a simple rectangle filled using the
        cursor color attribute.

        Parameters
        ----------
        surface:
            Surface representing the cursor.
        cursor:
            Cursor object.
        """
        if cursor.selected:
            surface.fill(self.selection_color)
        else:
            surface.fill(self.cursor_color)

    def draw_text(self, surface, text):
        """Default drawing method for text input box.

        Draw the text.

        Parameters
        ----------
        surface:
            Surface on which the text is drawn.
        text:
            Target text to be drawn.
        """
        if self.font_input_height != surface.get_height():
            # Resize font to fit the surface
            self.font_input = fit_font(self.font_name, surface.get_height())
            self.font_input_height = surface.get_height()

        surface.fill(self.background_input_color)
        surface.blit(self.font_input.render(text, 1,
                                            self.text_color[0]), (0, 0))

    def draw_key(self, surface, key):
        """Default drawing method for key.

        Draw the key accordingly to it type.

        Parameters
        ----------
        surface:
            Surface background should be drawn in.
        key:
            Target key to be drawn.
        """
        if isinstance(key, vkeys.VSpaceKey):
            self.draw_space_key(surface, key)
        elif isinstance(key, vkeys.VBackKey):
            self.draw_back_key(surface, key)
        elif isinstance(key, vkeys.VUppercaseKey):
            self.draw_uppercase_key(surface, key)
        elif isinstance(key, vkeys.VSpecialCharKey):
            self.draw_special_char_key(surface, key)
        else:
            self.draw_character_key(surface, key)

    def draw_character_key(self, surface, key, special=False):
        """Default drawing method for key.

        Key is drawn as a simple rectangle filled using this
        cell style background color attribute. Key value is printed
        into drawn cell using internal font.

        Parameters
        ----------
        surface:
            Surface key background should be drawn in.
        key:
            Target key to be drawn.
        special:
            Boolean flag that indicates if the drawn key should use
            special background color if available.
        """
        rect = surface.get_rect().inflate(-2, -2)
        if self.font_height != rect.height:
            # Resize font to fit the surface
            self.font = fit_font(self.font_name, rect.height)
            self.font_height = rect.height

        background_color = self.background_key_color[key.pressed]
        if special and self.background_special_key_color:
            background_color = self.background_special_key_color[key.pressed]

        text_color = self.text_color[key.pressed]
        if special and self.text_special_key_color and not key.pressed:
            # Key is not pressed, color according to activated state
            state = getattr(key, 'activated', key.pressed)
            text_color = self.text_special_key_color[state]

        surface.fill(self.background_color)
        draw_round_rect(surface, background_color, rect, 0.4)
        if key.selected:
            draw_round_rect(surface, self.selection_color, surface.get_rect(), 0.4, 1)

        text = self.font.render(str(key), 1, text_color)
        x = (key.rect.width - text.get_width()) // 2
        y = (key.rect.height - text.get_height()) // 2
        surface.blit(text, (x, y))

    def draw_space_key(self, surface, key):
        """Default drawing method space key.

        Key is drawn as a simple rectangle filled using this
        cell style background color attribute. Key value is printed
        into drawn cell using internal font.

        Parameters
        ----------
        surface:
            Surface background should be drawn in.
        key:
            Target key to be drawn.
        """
        self.draw_character_key(surface, key, False)

    def draw_back_key(self, surface, key):
        """Default drawing method for back key. Drawn as character key.

        Parameters
        ----------
        surface:
            Surface background should be drawn in.
        key:
            Target key to be drawn.
        """
        self.draw_character_key(surface, key, True)

    def draw_uppercase_key(self, surface, key):
        """Default drawing method for uppercase key. Drawn as character key.

        Parameters
        ----------
        surface:
            Surface background should be drawn in.
        key:
            Target key to be drawn.
        """
        self.draw_character_key(surface, key, True)

    def draw_special_char_key(self, surface, key):
        """Default drawing method for special char key.
        Drawn as character key.

        Parameters
        ----------
        surface:
            Surface background should be drawn in.
        key:
            Target key to be drawn.
        """
        self.draw_character_key(surface, key, True)


# Default keyboard renderer
VKeyboardRenderer.DEFAULT = VKeyboardRenderer(
    osp.join(osp.dirname(__file__), 'DejaVuSans.ttf'),
    text_color=((0, 0, 0), (255, 255, 255)),
    cursor_color=(0, 0, 0),
    selection_color=(0, 0, 200),
    background_color=(255, 255, 255),
    background_key_color=((255, 255, 255), (0, 0, 0)),
    background_input_color=(220, 220, 220),
    background_special_key_color=((180, 180, 180), (0, 0, 0))
)

VKeyboardRenderer.DARK = VKeyboardRenderer(
    osp.join(osp.dirname(__file__), 'DejaVuSans.ttf'),
    text_color=((182, 183, 184), (255, 255, 255)),
    cursor_color=(255, 255, 255),
    selection_color=(124, 183, 62),
    background_color=(0, 0, 0),
    background_key_color=((59, 56, 54), (47, 48, 51)),
    background_input_color=(80, 80, 80),
    text_special_key_color=((182, 183, 184), (124, 183, 62)),
    background_special_key_color=((35, 33, 30), (47, 48, 51))
)



#!/usr/bin/env python
# coding: utf8

"""
Text box to display the current text. The mouse events are
supported to move the cursor at the desired place.
"""

import pygame  # pylint: disable=import-error

from .vrenderers import VKeyboardRenderer


class VBackground(pygame.sprite.DirtySprite):
    """Background of the text input box. It is used to create
    borders by making its size a litle bit larger than the
    lines width and the sum of lines heights.
    """

    def __init__(self, size, renderer):
        """Default constructor.

        Parameters
        ----------
        size:
            Size tuple (width, height) of the background.
        renderer:
            Renderer used to draw the background.
        """
        super(VBackground, self).__init__()
        self.renderer = renderer
        self.image = pygame.Surface(size, pygame.SRCALPHA, 32)
        self.rect = pygame.Rect((0, 0), size)

        self.renderer.draw_background(self.image)

    def set_rect(self, x, y, width, height):
        """Set the background absolute position and size.

        Parameters
        ----------
        x:
            Position x.
        y:
            Position y.
        width:
            Background width.
        height:
            Background height.
        """
        if self.rect.topleft != (x, y):
            self.rect.topleft = (x, y)
            self.dirty = 1
        if self.rect.size != (width, height):
            self.rect.size = (width, height)
            self.image = pygame.Surface((width, height), pygame.SRCALPHA, 32)
            self.renderer.draw_background(self.image)
            self.dirty = 1


class VCursor(pygame.sprite.DirtySprite):
    """Handles the cursor.

    The ``index`` represente the absolute position which is the number
    of characters before it, all lines included.
    """

    def __init__(self, size, renderer):
        """Default constructor.

        Parameters
        ----------
        size:
            Size tuple (width, height) of the cursor.
        renderer:
            Renderer used to draw the cursor.
        """
        super(VCursor, self).__init__()
        self.renderer = renderer
        self.image = pygame.Surface(size, pygame.SRCALPHA, 32)
        self.rect = self.image.get_rect()
        self.index = 0
        self.selected = 0

        # Blink management
        self.clock = pygame.time.Clock()
        self.switch_ms = 400
        self.switch_counter = 0

        self.renderer.draw_cursor(self.image, self)

    def set_position(self, position):
        """Set the cursor absolute position.

        Parameters
        ----------
        position:
            Position tuple (x, y).
        """
        if self.rect.topleft != position:
            self.rect.topleft = position
            self.dirty = 1

    def set_size(self, width, height):
        """Set the cursor size.

        Parameters
        ----------
        width:
            Background width.
        height:
            Background height.
        """
        if self.rect.size != (width, height):
            self.rect.size = (width, height)
            self.image = pygame.Surface(self.rect.size, pygame.SRCALPHA, 32)
            self.renderer.draw_cursor(self.image, self)
            self.dirty = 1

    def set_index(self, index):
        """Move the cursor at the given index.

        Parameters
        ----------
        index:
            Absolute (sum all lines) cursor index.
        """
        if index != self.index:
            self.index = index
            self.dirty = 1

    def set_selected(self, state):
        """Set the key selection state (1 for selected else 0)
        and redraws it.

        Parameters
        ----------
        state:
            New key state.
        """
        if self.selected != int(state):
            self.selected = int(state)
            self.renderer.draw_cursor(self.image, self)
            self.dirty = 1

    def update(self, events):
        """Toggle visibility of the cursor."""
        self.clock.tick()
        self.switch_counter += self.clock.get_time()
        if self.switch_counter >= self.switch_ms:
            self.switch_counter %= self.switch_ms
            self.visible = int(not self.visible)
            self.dirty = 1


class VLine(pygame.sprite.DirtySprite):
    """Handles a line of text. A line can be fed until the text
    width reaches the line width.

    By default, when the line is empty, its ``visible`` attribute
    is set to 0 to hide the line.
    """

    def __init__(self, size, renderer, always_visible=False):
        """Default constructor.

        Parameters
        ----------
        size:
            Size tuple (width, height) of the line.
        renderer:
            Renderer used to draw the line.
        always_visible:
            If True, the line will be never hidden even if it is empty.
        """
        super(VLine, self).__init__()
        self.renderer = renderer
        self.image = pygame.Surface(size, pygame.SRCALPHA, 32)
        self.rect = pygame.Rect((0, 0), size)
        self.text = ''
        self.full = False
        self.always_visible = always_visible

        self.renderer.draw_text(self.image, '')

    def __len__(self):
        """Return the number of characters in the line."""
        return len(self.text)

    def set_position(self, position):
        """Set the line absolute position.

        Parameters
        ----------
        position:
            Position tuple (x, y).
        """
        if self.rect.topleft != position:
            self.rect.topleft = position
            self.dirty = 1

    def set_size(self, width, height):
        """Set the line size.

        Parameters
        ----------
        width:
            Background width.
        height:
            Background height.
        """
        if self.rect.size != (width, height):
            self.rect.size = (width, height)
            self.image = pygame.Surface(self.rect.size, pygame.SRCALPHA, 32)
            self.renderer.draw_text(self.image, self.text)
            self.dirty = 1

    def clear(self):
        """Clear the current text."""
        if self.text:
            self.text = ''
            self.full = False
            self.renderer.draw_text(self.image, '')
            if not self.always_visible:
                self.visible = 0
            else:
                self.dirty = 1
        return self.text

    def feed(self, text):
        """Feed the line with the given text. The current text is
        cleared if an empty string is given.

        Parameters
        ----------
        text:
            Text to feed in.

        Returns
        -------
        remain:
            Return the remaining text if the line is full.
        """
        if not text:
            return self.clear()
        elif self.text:
            if text.startswith(self.text):
                if self.full:
                    return text[len(self.text):]
            else:
                self.text = ''

        self.text, _ = self.renderer.truncate(text,
                                              self.rect.width,
                                              len(self.text))
        if text[len(self.text):]:
            self.full = True
        else:
            self.full = False
        self.dirty = 1
        self.visible = 1  # Show line
        self.renderer.draw_text(self.image, self.text)
        return text[len(self.text):]


class VTextInput(object):
    """Handles the text input box.
    """

    def __init__(self,
                 position,
                 size,
                 border=2,
                 renderer=VKeyboardRenderer.DEFAULT):
        """Default constructor.

        Parameters
        ----------
        position:
            Position tuple (x, y)
        size:
            Size tuple (width, height) of the text input.
        border:
            Border thickness.
        renderer:
            Text input renderer instance, using VTextInputRenderer.DEFAULT
            if not specified.
        """
        self.state = 0
        self.selected = 0
        self.position = position
        self.size = size  # One ligne size
        self.text = ''
        self.text_margin = border
        self.renderer = renderer

        # Define background sprites
        self.eraser = None
        self.background = VBackground(size, renderer)
        self.sprites = pygame.sprite.LayeredDirty(self.background)

        # Initialize first line
        line = VLine((self.size[0] - 2 * self.text_margin,
                      self.size[1]), renderer, True)
        self.sprites.add(line, layer=1)

        # Initialize cursor
        self.cursor = VCursor((2, self.size[1] - self.text_margin * 2), renderer)
        self.sprites.add(self.cursor, layer=2)

        self.disable()

        self.set_line_rect(*self.position + self.size)

    def set_eraser(self, surface):
        """Setup the surface used to hide/clear the text input.
        """
        self.eraser = surface.copy()
        self.sprites.clear(surface, self.eraser)

    def enable(self):
        """Set this text input as active."""
        self.state = 1
        self.cursor.visible = 1
        self.background.visible = 1
        self.sprites.get_sprites_from_layer(1)[0].visible = 1

    def is_enabled(self):
        """Return True if this keyboard is active."""
        return self.state == 1

    def disable(self):
        """Set this text input as non active."""
        self.state = 0   # Disabled by default
        self.cursor.visible = 0
        self.background.visible = 0
        for line in self.sprites.get_sprites_from_layer(1):
            line.visible = 0

    def set_selected(self, state):
        """Set the input box selection state (1 for selected else 0)
        and redraws it.

        Parameters
        ----------
        state:
            New key state.
        """
        self.selected = int(state)
        self.cursor.set_selected(state)

    def get_rect(self):
        """Return text input rect."""
        return self.background.rect

    def set_line_rect(self, x, y, width, height):
        """Set the input text (one line) absolute position and size.

        Parameters
        ----------
        x:
            Position x.
        y:
            Position y.
        width:
            Background width.
        height:
            Background height.
        """
        self.size = (width, height)
        self.position = (x, y)
        self.background.set_rect(self.position[0],
                                 self.position[1] - 2 * self.text_margin,
                                 width,
                                 height + 2 * self.text_margin)

        for line in self.sprites.get_sprites_from_layer(1):
            line.clear()
            line.set_position((self.position[0] + self.text_margin,
                               self.position[1] - self.text_margin))
            line.set_size(self.size[0] - 2 * self.text_margin, self.size[1])
        self.update_lines()

        self.cursor.set_position((self.position[0] + self.text_margin,
                                  self.position[1]))
        self.cursor.set_size(2, self.size[1] - self.text_margin * 2)

        # Setup new the surface where to draw
        clip_rect = pygame.Rect(self.position[0], 0,
                                self.background.rect.width,
                                self.background.rect.bottom)
        if self.sprites.get_clip() != clip_rect:
            # Changing the clipping area will force update of all
            # sprites without using "dirty mechanism"
            self.sprites.set_clip(clip_rect)

    def draw(self, surface, force):
        """Draw the text input box.

        Parameters
        ----------
        surface:
            Surface on which the VTextInput is drawn.
        force:
            Force the drawing of the entire surface (time consuming).
        """
        # Check if surface has been resized
        if self.eraser and surface.get_rect() != self.eraser.get_rect():
            force = True  # To force creating new eraser

        # Setup eraser
        if not self.eraser or force:
            self.set_eraser(surface)

        if force:
            self.sprites.repaint_rect(self.background.rect)
        return self.sprites.draw(surface)

    def update(self, events):
        """Pygame events processing callback method.

        Parameters
        ----------
        events:
            List of events to process.
        """
        if self.state == 1:
            self.sprites.update(events)
            for event in events:
                if event.type == pygame.KEYUP and self.cursor.selected:
                    if event.key == pygame.K_LEFT:
                        self.increment_cursor(-1)
                    elif event.key == pygame.K_RIGHT:
                        self.increment_cursor(1)
                    elif event.key == pygame.K_HOME:
                        self.cursor.index = 0
                        self.increment_cursor(0)
                    elif event.key == pygame.K_END:
                        self.cursor.index = 0
                        self.increment_cursor(len(self.text))
                if event.type == pygame.MOUSEBUTTONDOWN\
                        and event.button in (1, 2, 3):
                    # Don't consider the mouse wheel (button 4 & 5):
                    self.set_cursor(event.pos)
                if event.type == pygame.FINGERDOWN:
                    display_size = pygame.display.get_surface().get_size()
                    finger_pos = (event.x * display_size[0], event.y * display_size[1])
                    self.set_cursor(finger_pos)

    def update_lines(self):
        """Update lines content with the current text."""
        if self.state == 1:
            remain = self.text

            # Update existing line with text
            for line in self.sprites.get_sprites_from_layer(1):
                remain = line.feed(remain)

            # Create new lines if necessary
            while remain:
                line = VLine((self.size[0] - 2 * self.text_margin,
                              self.size[1]), self.renderer)
                self.sprites.add(line, layer=1)
                remain = line.feed(remain)

            # Update lines positions
            i = 0
            for line in reversed(self.sprites.get_sprites_from_layer(1)):
                if line.visible:
                    x = self.position[0] + self.text_margin
                    y = self.position[1] - i * self.size[1] - self.text_margin
                    line.set_position((x, y))
                    i += 1

            self.background.set_rect(self.position[0],
                                     line.rect.y - self.text_margin,
                                     self.size[0],
                                     i * self.size[1] + 2 * self.text_margin)

    def set_text(self, text):
        """Overwrite the current text with the given one. The cursor is
        moved at the end of the text.

        Parameters
        ----------
        text:
            New text.
        """
        self.text = text
        self.update_lines()
        self.cursor.index = 0
        self.increment_cursor(len(text))

    def add_at_cursor(self, text):
        """Add a text whereever the cursor is currently located.

        Parameters
        ----------
        text:
            Single char or text to append.
        """
        if self.cursor.index < len(self.text):
            # Inserting in the text
            prefix = self.text[:self.cursor.index]
            suffix = self.text[self.cursor.index:]
            self.text = prefix + text + suffix
        else:
            self.text += text
        self.update_lines()
        self.increment_cursor(1)

    def delete_at_cursor(self):
        """Delete a character before the cursor position."""
        if self.cursor.index == 0:
            return
        prefix = self.text[:self.cursor.index - 1]
        suffix = self.text[self.cursor.index:]
        self.text = prefix + suffix
        self.update_lines()
        self.increment_cursor(-1)

    def increment_cursor(self, step):
        """Move the cursor of one or more steps (but not beyond the
        text length).

        Parameters
        ----------
        step:
            From how many characters the cursor shall move.
        """
        pos = max(0, self.cursor.index + step)
        self.cursor.set_index(min(pos, len(self.text)))

        # Update cursor position
        chars_counter = 0
        for line in self.sprites.get_sprites_from_layer(1):
            if chars_counter + len(line) >= self.cursor.index:
                idx = self.cursor.index - chars_counter
                x = self.text_margin + self.renderer.get_text_width(
                    line.text[:idx])
                self.cursor.set_position((x, line.rect.y + self.text_margin))
                break
            chars_counter += len(line)

    def set_cursor(self, position):
        """Move cursor to char nearest position (x, y).

        Parameters
        ----------
        position:
            Absolute position (x, y) on the surface.
        """
        for collide in self.sprites.get_sprites_at(position):
            if isinstance(collide, VLine):
                text, width = self.renderer.truncate(
                    collide.text,
                    position[0] - collide.rect.left,
                    nearest=True)
                self.cursor.set_position((
                    width + self.text_margin,
                    collide.rect.y + self.text_margin))
                chars_counter = 0
                for line in self.sprites.get_sprites_from_layer(1):
                    if line == collide:
                        self.cursor.set_index(chars_counter + len(text))
                        return
                    chars_counter += len(line)



#!/usr/bin/env python
# coding: utf8

"""
Module for keys definitions. It contains the following classes:

- `VKey`           : base class for all keys
- `VSpaceKey`      : *space* key definition
- `VBackKey`       : *delete* key definition

- `VActionKey`     : base class for key without effect on text
- `VUppercaseKey`  : *shift* key definition
- `VSpecialCharKey`: change the keyboard layout
"""

import pygame  # pylint: disable=import-error


class VKey(pygame.sprite.DirtySprite):
    """
    Simple key holder class.

    Holds key information (its value), as its state, size / position.
    State attributes with their default values:

    pressed = 0
        If set to 0, the key is released.
        If set to 1, the key is pressed.

    selected = 0
        If set to 0, the key is selectable but not selected.
        If set to 1, the key is selected.
    """

    def __init__(self, value, symbol=None):
        """Default key constructor.

        Parameters
        ----------
        value:
            Value of this key.
        symbol:
            Visual representation of the key displayed to the screen
            (equal to the value if not given).
        """
        super(VKey, self).__init__()
        self.pressed = 0
        self.selected = 0
        self.value = value
        self.symbol = symbol
        self.rect = pygame.Rect((0, 0), (10, 10))
        self.image = pygame.Surface(self.rect.size, pygame.SRCALPHA, 32)
        self.renderer = None
        self.pressed_key = None

    def __str__(self):
        """Key representation when using str() or print()"""
        if self.symbol:
            return self.symbol
        return self.value

    def set_position(self, x, y):
        """Set the key position.

        Parameters
        ----------
        x:
            Position x.
        y:
            Position y.
        """
        if self.rect.topleft != (x, y):
            self.rect.topleft = (x, y)
            self.dirty = 1

    def set_size(self, width, height):
        """Set the key size.

        Parameters
        ----------
        width:
            Background width.
        height:
            Background height.
        """
        if self.rect.size != (width, height):
            self.rect.size = (width, height)
            self.image = pygame.Surface(self.rect.size, pygame.SRCALPHA, 32)
            self.renderer.draw_key(self.image, self)
            self.dirty = 1

    def set_uppercase(self, uppercase):
        """Set key uppercase state and redraws it.

        Parameters
        ----------
        uppercase:
            True if uppercase, False otherwise.
        """
        if uppercase:
            new_value = self.value.upper()
        else:
            new_value = self.value.lower()
        if new_value != self.value:
            self.value = new_value
            self.renderer.draw_key(self.image, self)
            self.dirty = 1

    def set_pressed(self, state):
        """Set the key pressed state (1 for pressed 0 for released)
        and redraws it.

        Parameters
        ----------
        state:
            New key state.
        """
        if self.pressed != int(state):
            self.pressed = int(state)
            self.renderer.draw_key(self.image, self)
            self.dirty = 1

    def set_selected(self, state):
        """Set the key selection state (1 for selected else 0)
        and redraws it.

        Parameters
        ----------
        state:
            New key state.
        """
        if self.selected != int(state):
            self.selected = int(state)
            self.renderer.draw_key(self.image, self)
            self.dirty = 1

    def update(self, events):
        """Pygame events processing callback method.

        Parameters
        ----------
        events:
            List of events to process.
        """
        for event in events:
            if event.type == pygame.MOUSEBUTTONDOWN\
                    and event.button in (1, 2, 3):
                # Don't consider the mouse wheel (button 4 & 5):
                if self.rect.collidepoint(event.pos):
                    self.set_pressed(1)
            elif event.type == pygame.MOUSEBUTTONUP\
                    and event.button in (1, 2, 3):
                # Don't consider the mouse wheel (button 4 & 5):
                self.set_pressed(0)
            elif event.type == pygame.FINGERDOWN:
                display_size = pygame.display.get_surface().get_size()
                finger_pos = (event.x * display_size[0], event.y * display_size[1])
                if self.rect.collidepoint(finger_pos):
                    self.set_pressed(1)
            elif event.type == pygame.FINGERUP:
                self.set_pressed(0)
            elif event.type == pygame.KEYDOWN:
                if event.unicode and event.unicode == self.value:
                    self.set_pressed(1)
                    self.pressed_key = event.key
                elif event.key == self.value:
                    self.set_pressed(1)
                    self.pressed_key = event.key
                elif event.key == pygame.K_RETURN and self.selected:
                    self.set_pressed(1)
                    self.pressed_key = event.key
            elif event.type == pygame.KEYUP and self.pressed_key is not None:
                self.set_pressed(0)
                self.pressed_key = None
            elif event.type == pygame.JOYBUTTONDOWN and event.button == 0\
                    and self.selected:  # Select button pressed
                self.set_pressed(1)
            elif event.type == pygame.JOYBUTTONUP and event.button == 0\
                    and self.selected:  # Select button released
                self.set_pressed(0)

    def update_buffer(self, string):
        """Text update method.

        Aims to be called internally when a key collision has been detected.
        Updates and returns the given buffer using this key value.

        Parameters
        ----------
        string:
            Buffer to be updated.

        Returns
        -------
        string:
            Updated buffer value.
        """
        return string + self.value


class VSpaceKey(VKey):
    """Custom key for spacebar. """

    def __init__(self, length):
        """Default constructor.

        Parameters
        ----------
        length:
            Key length.
        """
        VKey.__init__(self, ' ', u'space')
        self.length = length

    def set_uppercase(self, uppercase):
        """Nothing to do on upper case action."""
        pass

    def set_size(self, width, height):
        """Sets the size of this key.

        Parameters
        ----------
        width:
            Background width.
        height:
            Background height.
        """
        super(VSpaceKey, self).set_size(width * self.length, height)


class VBackKey(VKey):
    """Custom key for back. """

    def __init__(self):
        """Default constructor. """
        VKey.__init__(self, u'\x7f', u'\u2190')

    def set_uppercase(self, uppercase):
        """Nothing to do on upper case action."""
        pass

    def update_buffer(self, string):
        """Text update method. Removes last character.

        Parameters
        ----------
        string:
            Buffer to be updated.

        Returns
        -------
        string:
            Updated buffer value.
        """
        return string[:-1]


class VActionKey(VKey):
    """
    A VActionKey is a key that trigger an action
    rather than updating the buffer when pressed.
    """

    def __init__(self, action, state_holder, symbol, activated_symbol):
        """Default constructor.

        Parameters
        ----------
        action:
            Delegate action called when this key is pressed.
        state_holder:
            Holder for this key state (activated or not).
        """
        super(VActionKey, self).__init__('', symbol)
        self.action = action
        self.state_holder = state_holder
        self.activated_symbol = activated_symbol
        self.activated = False

    def __str__(self):
        """Key representation when using str() or print()"""
        if self.is_activated():
            self.activated = True
            return self.activated_symbol
        return self.symbol

    def update(self, events):
        """Check if state holder has changed."""
        super(VActionKey, self).update(events)
        if self.activated != self.is_activated() and not self.dirty:
            self.activated = self.is_activated()
            self.renderer.draw_key(self.image, self)
            self.dirty = 1

    def set_uppercase(self, uppercase):
        """Nothing to do on upper case action."""
        pass

    def set_pressed(self, state):
        """Nothing to do on upper case action."""
        prev_state = self.pressed
        super(VActionKey, self).set_pressed(state)
        if prev_state != self.pressed and self.pressed == 0:
            # The key is getting unpressed
            self.action()

    def is_activated(self):
        """Indicates if this key is activated.

        Returns
        -------
        is_activated: bool
            True if activated, False otherwise.
        """
        raise NotImplementedError(
            "Method 'is_activated' have to be overwritten")

    def update_buffer(self, string):
        """Do not update text but trigger the delegate action.

        Parameters
        ----------
        string:
            Not used, just to match parent interface.

        Returns
        -------
        string:
            Buffer provided as parameter.
        """
        return string


class VUppercaseKey(VActionKey):
    """Action key for the uppercase switch. """

    def __init__(self, action, state_holder):
        super(VUppercaseKey, self).__init__(action,
                                            state_holder,
                                            u'\u21e7',
                                            u'\u21ea')
        self.value = pygame.K_LSHIFT

    def is_activated(self):
        """Indicates if this key is activated.

        Returns
        -------
        is_activated: bool
            True if activated, False otherwise.
        """
        return self.state_holder.uppercase


class VSpecialCharKey(VActionKey):
    """Action key for the special char switch. """

    def __init__(self, action, state_holder):
        super(VSpecialCharKey, self).__init__(action,
                                              state_holder,
                                              u'#',
                                              u'Ab')

    def is_activated(self):
        """Indicates if this key is activated.

        Returns
        -------
        is_activated: bool
            True if activated, False otherwise.
        """
        return self.state_holder.special_char



#!/usr/bin/env python
# coding: utf8

"""
Visual keyboard for Pygame engine. Aims to be easy to use
as highly customizable as well.

``VKeyboard`` only require a pygame surface to be displayed
on and a text consumer function, as in the following example :

```python
from pygame_vkeyboard import VKeyboard, VKeyboardLayout

# Initializes your window object or surface your want
# vkeyboard to be displayed on top of.
surface = ...

def consume(text):
    print(repr('Current text : %s' % text))

# Initializes and activates vkeyboard
layout = VKeyboardLayout(VKeyboardLayout.AZERTY)
keyboard = VKeyboard(window, consumer, layout)
keyboard.enable()
```
"""

import logging
import pygame  # pylint: disable=import-error

from . import vkeys
from .vrenderers import VKeyboardRenderer
from .vtextinput import VTextInput, VBackground


LOGGER = logging.getLogger(__name__)

# Joystick controls
JOYHAT_UP = (0, 1)
JOYHAT_LEFT = (-1, 0)
JOYHAT_RIGHT = (1, 0)
JOYHAT_DOWN = (0, -1)


def synchronize_layouts(surface_size, *layouts):
    """Synchronizes given layouts by normalizing height by using
    max height of given layouts to avoid transistion dirty effects.

    Parameters
    ----------
    surface_size:
        Target surface size on which layout will be displayed.
    layouts:
        All layouts to synchronize
    """
    for layout in layouts:
        layout.configure_bound(surface_size)

    layout_ref = min(layouts, key=lambda l: l.key_size)

    for layout in layouts:
        if layout.key_size != layout_ref.key_size:
            logging.warning(
                'Normalizing layout%s key size to %spx (from layout%s)',
                layouts.index(layout),
                layout_ref.key_size,
                layouts.index(layout_ref))
            layout.key_size = layout_ref.key_size
        if layout.size != layout_ref.size:
            logging.warning(
                'Normalizing layout%s size to %s*%spx (from layout%s)',
                layouts.index(layout),
                layout_ref.size[0], layout_ref.size[1],
                layouts.index(layout_ref))

        # Compute all internal values of the layout
        layout.set_size(layout_ref.size, surface_size)


class VKeyRow(object):
    """A VKeyRow defines a keyboard row which is composed of a list of
    VKey.

    This class aims to be created internally after parsing a keyboard
    layout model. It is used to optimize collision detection, by first
    checking row collision, then internal row key detection.
    """

    def __init__(self):
        """Default row constructor. """
        self.keys = []
        self.height = 0
        self.position = (0, 0)
        self.space = None

    def add_key(self, key, first=False):
        """Adds the given key to this row.

        Parameters
        ----------
        key:
            Key to be added to this row.
        first:
            Flag that indicates if key is added at the beginning or at the end.
        """
        if first:
            self.keys = [key] + self.keys
        else:
            self.keys.append(key)
        if isinstance(key, vkeys.VSpaceKey):
            self.space = key

    def set_size(self, position, size, padding):
        """Row size setter. The size correspond to the row height, since the
        row width is constraint to the surface width the associated keyboard
        belongs. Once size is settled, the size for each child keys is
        associated.

        Parameters
        ----------
        position:
            Position of this row.
        size:
            Size of the row (height)
        padding:
            Padding between key.
        """
        self.height = size
        self.position = position
        x = position[0]
        for key in self.keys:
            key.set_size(size, size)
            key.set_position(x, position[1])
            x += padding + key.rect.width

    def __len__(self):
        """len() operator overload.

        Returns
        -------
        len:
            Number of keys thi row contains.
        """
        return len(self.keys)


class VKeyboardLayout(object):
    """Keyboard layout class.

    A keyboard layout is built using layout model which consists in an
    list of supported character. Such list item as simple string containing
    characters assigned to a row.

    An erasing key is inserted automatically to the first row.

    If `allow_uppercase` flag is `True`, then an upper case key will be
    inserted at the beginning of the second row.

    If `allow_special_chars` flag is `True`, then an special
    characters / number key will be inserted at the beginning of the third row.
    Pressing this key will switch the associated keyboard current layout.
    """

    # AZERTY Layout.
    AZERTY = ['1234567890', 'azertyuiop', 'qsdfghjklm', 'wxcvbn']

    # QWERTY Layout.
    QWERTY = ['1234567890', 'qwertyuiop', 'asdfghjkl', 'zxcvbnm']

    # Number only layout.
    NUMBER = ['123', '456', '789', '0']

    # TODO : Insert special characters layout which include number.
    SPECIAL = [u'&é"\'(§è!çà)', u'°_-^$¨*ù`%£', u',;:=?.@+<>#', u'€[]{}/\\|']

    def __init__(self,
                 model,
                 key_size=None,
                 padding=5,
                 height_ratio=None,
                 allow_uppercase=True,
                 allow_special_chars=True,
                 allow_space=True):
        """Default constructor. Initializes layout rows.

        Parameters
        ----------
        model:
            Layout model to use.
        key_size:
            Size of the key, dynamically computed if not specified.
        padding:
            Padding between key (work horizontally as vertically).
        height_ratio:
            Ratio (0.2 to 1) of the surface height to recover, dynamically
            computed if not specified.
        allow_uppercase:
            Boolean flag that indicates usage of upper case switching key.
        allow_special_chars:
            Boolean flag that indicates usage of special char switching key.
        allow_space:
            Boolean flag that indicates usage of space bar.

        Raises
        ------
        ValueError
            If the layout model is empty.
        """
        self.position = None
        self.size = None
        self.rows = []
        self.sprites = pygame.sprite.LayeredDirty()
        self.padding = padding
        self.height_ratio = height_ratio
        self.selection = None
        self.allow_space = allow_space
        self.allow_uppercase = allow_uppercase
        self.allow_special_chars = allow_special_chars
        for model_row in model:
            row = VKeyRow()
            for value in model_row:
                key = vkeys.VKey(value)
                row.add_key(key)
                self.sprites.add(key, layer=1)
            self.rows.append(row)
        self.max_length = len(max(self.rows, key=len))
        if self.max_length == 0:
            raise ValueError('Empty layout model provided')
        if height_ratio is not None and (height_ratio < 0.2 or height_ratio > 1):
            raise ValueError('Surface height ratio shall be from 0.2 to 1')

        self._key_size = key_size
        self._key_size_computed = None

    @property
    def key_size(self):
        return self._key_size or self._key_size_computed

    @key_size.setter
    def key_size(self, key_size):
        self._key_size = key_size

    def hide(self):
        """Hide all keys."""
        for sprite in self.sprites:
            sprite.visible = 0

    def show(self):
        """Show all keys."""
        for sprite in self.sprites:
            sprite.visible = 1

    def configure_renderer(self, renderer):
        """Configure the keys with the given renderer.

        Parameters
        ----------
        renderer:
            Renderer instance this layout uses.
        """
        for key in self.sprites.get_sprites_from_layer(1):
            key.renderer = renderer

    def configure_special_keys(self, keyboard):
        """Configures specials key if needed.

        Parameters
        ----------
        keyboard:
            Keyboard instance this layout belong.
        """
        special_row = VKeyRow()
        max_length = self.max_length
        i = len(self.rows) - 1
        current_row = self.rows[i]

        # Create special keys list
        special_keys = [vkeys.VBackKey()]
        if self.allow_uppercase:
            special_keys.append(
                vkeys.VUppercaseKey(keyboard.on_uppercase, keyboard))
        if self.allow_special_chars:
            special_keys.append(
                vkeys.VSpecialCharKey(keyboard.on_special_char, keyboard))
        self.sprites.add(*special_keys, layer=1)

        # Dispatch special keys in the layout
        while special_keys:
            first = False
            while special_keys and len(current_row) < max_length:
                current_row.add_key(special_keys.pop(0), first=first)
                first = not first
            if i > 0:
                i -= 1
                current_row = self.rows[i]
            else:
                break
        if self.allow_space:
            space_length = len(current_row) - len(special_keys)
            special_row.add_key(vkeys.VSpaceKey(space_length))
            self.sprites.add(special_row.space, layer=1)
        first = True

        # Adding left to the special bar.
        while special_keys:
            special_row.add_key(special_keys.pop(0), first=first)
            first = not first

        if special_row:
            self.rows.append(special_row)

    def configure_bound(self, surface_size):
        """Compute keyboard bound regarding of this layout. If `key_size` is
        `None`, then it will compute it regarding of the given surface_size.

        Parameters
        ----------
        surface_size:
            Size of the surface this layout will be rendered on.
        """
        nb_rows = len(self.rows)
        if self._key_size is None:
            self._key_size_computed = int(
                (surface_size[0] - (self.padding * (self.max_length + 1)))
                / self.max_length)

        height = self.key_size * nb_rows + self.padding * (nb_rows + 1)
        if height > surface_size[1] * (self.height_ratio or 0.5):
            self._key_size_computed = int((surface_size[1] * (self.height_ratio or 0.5)
                                           - (self.padding * (nb_rows + 1))) / nb_rows)
            height = self.key_size * nb_rows + self.padding * (nb_rows + 1)
            if self._key_size:
                self._key_size = self._key_size_computed
                LOGGER.warning('Computed layout height outbound target surface,'
                               ' reducing key_size to %spx', self.key_size)
        elif self.height_ratio is not None:
            height = surface_size[1] * self.height_ratio
        self.set_size((surface_size[0], int(height)), surface_size)

    def set_size(self, size, surface_size):
        """Sets the size of this layout, and updates
        position, and rows accordingly.

        Parameters
        ----------
        size:
            Size of this layout.
        surface_size:
            Target surface size on which layout will be displayed.
        """
        self.size = size
        self.position = (0, surface_size[1] - self.size[1])

        y = self.position[1] + (self.size[1] - len(self.rows) * self.key_size
                                - (len(self.rows) + 1) * self.padding) // 2
        y += self.padding
        for row in self.rows:
            nb_keys = len(row)
            width = (nb_keys * self.key_size) + ((nb_keys + 1) * self.padding)
            x = (surface_size[0] - width) // 2 + self.padding
            if row.space:
                x -= ((row.space.length - 1) * self.key_size) / 2
            row.set_size((x, y), self.key_size, self.padding)
            y += self.padding + self.key_size

    def set_uppercase(self, uppercase):
        """Sets layout uppercase state.

        Parameters
        ----------
        uppercase:
            True if uppercase, False otherwise.
        """
        for key in self.sprites.get_sprites_from_layer(1):
            key.set_uppercase(uppercase)

    def get_key(self, value):
        """Retrieve if any key with the given value

        Parameters
        ----------
        value:
            Value to find among keys.

        Returns
        -------
        key:
            The located key if any with the given value, None otherwise.
        """
        for key in self.sprites.get_sprites_from_layer(1):
            if key.value == value:
                return key
        return None

    def get_key_at(self, position):
        """Retrieve if any key is located at the given position

        Parameters
        ----------
        position:
            Position to check key at.

        Returns
        -------
        key:
            The located key if any at the given position, None otherwise.
        """
        for sprite in self.sprites.get_sprites_at(position):
            if isinstance(sprite, vkeys.VKey):
                return sprite
        return None

    def get_key_closest(self, key, loop_row=True, loop_col=True):
        """Retrieve the keys closest to the given one. It returns
        a dictionary with closest keys and their position relative
        to the given key (diff row, diff column).

        Parameters
        ----------
        key:
            The key to locate.
        loop_row:
            Loop to the start/end of the row when right/left overflow.
        loop_col:
            Loop to the start/end of the column when bottom/top overflow.

        Returns
        -------
        keys_dict:
            The located keys on the left, top, right, bottom.
        """
        center = (0, 0)
        left = (0, -1)
        top = (-1, 0)
        right = (0, 1)
        bottom = (1, 0)
        keys_dict = {center: key,
                     left: None, top: None, right: None, bottom: None}

        for row_index, r in enumerate(self.rows):
            for key_index, k in enumerate(r.keys):
                if key == k:
                    if key_index - 1 >= 0 or loop_row:
                        keys_dict[left] = r.keys[(key_index - 1) % len(r.keys)]
                    if key_index + 1 < len(r.keys) or loop_row:
                        keys_dict[right] = r.keys[(key_index + 1) % len(r.keys)]

                    if row_index - 1 >= 0 or loop_col:
                        prev_row = self.rows[(row_index - 1) % len(self.rows)]
                        keys_dict[top] = prev_row\
                            .keys[min(key_index, len(prev_row) - 1)]
                    if row_index + 1 < len(self.rows) or loop_col:
                        next_row = self.rows[(row_index + 1) % len(self.rows)]
                        keys_dict[bottom] = next_row\
                            .keys[min(key_index, len(next_row) - 1)]
        return keys_dict


class VKeyboard(object):
    """
    Virtual Keyboard class.

    A virtual keyboard consists in a VKeyboardLayout that acts as
    the keyboard model and a VKeyboardRenderer which is in charge
    of drawing keyboard component to screen.
    """

    def __init__(self,
                 surface,
                 text_consumer,
                 main_layout,
                 show_text=False,
                 joystick_navigation=False,
                 renderer=VKeyboardRenderer.DEFAULT,
                 special_char_layout=None):
        """ Default constructor.

        Parameters
        ----------
        surface:
            Surface this keyboard will be displayed at.
        text_consumer:
            Consumer that process text for each update.
        main_layout:
            First displayed layout of this keyboard.
        show_text:
            Display the current text in a text box.
        joystick_navigation:
            Handle joystick events and arrow keys.
        renderer:
            Keyboard renderer instance, using VKeyboardRenderer.DEFAULT
            if not specified.
        special_char_layout:
            Alternative layout to use, using VKeyboardLayout.SPECIAL
            if not specified.
        """
        self.surface = surface
        self.text_consumer = text_consumer
        self.renderer = renderer
        self.state = 1  # Enabled by default
        self.last_pressed = None
        self.uppercase = False
        self.special_char = False
        self.joystick_navigation = joystick_navigation

        # Setup background as a DirtySprite
        self.eraser = None
        self.background = VBackground(self.surface.get_rect().size,
                                      self.renderer)

        # Setup the layouts
        self.layout = main_layout
        self.layouts = [main_layout]
        if self.layout.allow_special_chars:
            if not special_char_layout:
                special_char_layout = VKeyboardLayout(
                    VKeyboardLayout.SPECIAL,
                    key_size=self.layout.key_size,
                    padding=self.layout.padding,
                    height_ratio=self.layout.height_ratio,
                    allow_uppercase=self.layout.allow_uppercase,
                    allow_special_chars=self.layout.allow_special_chars,
                    allow_space=self.layout.allow_space)
            self.layouts.append(special_char_layout)

        for layout in self.layouts:
            layout.configure_special_keys(self)
            layout.configure_renderer(self.renderer)
            layout.sprites.add(self.background, layer=0)

        # Setup the text input box
        self.show_text = show_text
        self.input = VTextInput((0, 0), (10, 10), renderer=self.renderer)

        self.resize(surface)

        if self.show_text:
            self.input.enable()

        # Setup de joystick
        if self.joystick_navigation:
            if not pygame.joystick.get_init():
                pygame.joystick.init()
            for i in range(pygame.joystick.get_count()):
                pygame.joystick.Joystick(i).init()

    def set_layout(self, layout):
        """Sets the layout this keyboard work with.
        Keyboard is invalidate by this action and redraw itself.

        Parameters
        ----------
        layout:
            Layout to set.
        """
        self.layout.hide()
        self.layout = layout
        self.layout.show()

    def set_eraser(self, surface):
        """Setup the surface used to hide/clear the keyboard.
        """
        self.eraser = surface.copy()
        for layout in self.layouts:
            layout.sprites.clear(surface, self.eraser)

    def resize(self, surface):
        """Resize the keyboard according to the surface size and the parameters
        of the layout(s).

        Parameters
        ----------
        surface:
            Surface this keyboard will be displayed at.
        """
        synchronize_layouts(surface.get_size(), *self.layouts)
        self.background.set_rect(*self.layout.position + self.layout.size)

        for layout in self.layouts:
            if layout.sprites.get_clip() != self.background.rect:
                # Changing the clipping area will force update of all
                # sprites without using "dirty mechanism"
                layout.sprites.set_clip(self.background.rect)

        self.input.set_line_rect(self.layout.position[0],
                                 self.layout.position[1] - self.layout.key_size,
                                 self.layout.size[0],
                                 self.layout.key_size)

    def set_text(self, text):
        """Set the current text in the internal buffer.

        Parameters
        ----------
        text:
            Text to be set.
        """
        self.input.set_text(text)

    def get_text(self):
        """Return the current text of the internal buffer."""
        return self.input.text

    def enable(self):
        """Set this keyboard as active."""
        self.state = 1
        self.layout.show()
        if self.show_text:
            self.input.enable()

    def is_enabled(self):
        """Return True if this keyboard is active."""
        return self.state == 1

    def disable(self):
        """Set this keyboard as non active."""
        self.state = 0
        self.layout.hide()
        self.input.disable()

    def get_rect(self):
        """Return keyboard rect."""
        rect = self.background.rect
        if self.show_text:
            rect = rect.union(self.input.get_rect())
        return rect

    def draw(self, surface=None, force=False):
        """Draw the virtual keyboard.

        This method is optimized to be called at each loop of the
        main application. It uses DirtySprite to update only parts
        of the screen that need to be refreshed.

        The first call to this method will setup the "eraser" surface that
        will be used to redraw dirty parts of the screen.

        The `force` parameter shall be used if the surface has been redrawn:
        it reset the eraser and redraw all sprites.

        Parameters
        ----------
        surface:
            Surface this keyboard will be displayed at.
        force:
            Force the drawing of the entire surface (time consuming).

        Returns
        -------
        rects:
            List of updated area.
        """
        surface = surface or self.surface

        # Check if surface has been resized
        if self.eraser and surface.get_rect() != self.eraser.get_rect():
            force = True  # To force creating new eraser
            self.resize(surface)

        # Setup eraser
        if not self.eraser or force:
            self.set_eraser(surface)

        rects = self.layout.sprites.draw(surface)
        rects += self.input.draw(surface, force)

        if force:
            self.layout.sprites.repaint_rect(self.background.rect)
        return rects

    def update(self, events):
        """Pygame events processing callback method.

        Parameters
        ----------
        events:
            List of events to process.
        """
        if self.state == 1:
            self.layout.sprites.update(events)
            self.input.update(events)

            for event in events:
                if event.type == pygame.MOUSEBUTTONDOWN\
                        and event.button in (1, 2, 3):
                    # Don't consider the mouse wheel (button 4 & 5)
                    key = self.layout.get_key_at(event.pos)
                    if key:
                        self.on_key_down(key)
                        self.on_select(0, 0, key)
                    elif self.input.get_rect().collidepoint(event.pos)\
                            and self.layout.selection:
                        self.layout.selection.set_selected(0)
                        self.layout.selection = None
                        self.input.set_selected(1)
                elif event.type == pygame.FINGERDOWN:
                    display_size = pygame.display.get_surface().get_size()
                    finger_pos = (event.x * display_size[0], event.y * display_size[1])
                    key = self.layout.get_key_at(finger_pos)
                    if key:
                        self.on_key_down(key)
                        self.on_select(0, 0, key)
                    elif self.input.get_rect().collidepoint(finger_pos)\
                            and self.layout.selection:
                        self.layout.selection.set_selected(0)
                        self.layout.selection = None
                        self.input.set_selected(1)
                elif event.type == pygame.KEYDOWN:
                    key = self.layout.get_key(event.unicode or event.key)
                    if key:
                        self.on_key_down(key)
                        self.on_select(0, 0, key)
                    elif event.key == pygame.K_LEFT and not self.input.selected:
                        self.on_select(0, -1)
                    elif event.key == pygame.K_UP:
                        self.on_select(-1, 0)
                    elif event.key == pygame.K_RIGHT and\
                            not self.input.selected:
                        self.on_select(0, 1)
                    elif event.key == pygame.K_DOWN:
                        self.on_select(1, 0)
                    elif event.key == pygame.K_RETURN and self.layout.selection:
                        self.on_key_down(self.layout.selection)
                elif event.type == pygame.JOYHATMOTION:
                    if event.value == JOYHAT_LEFT and\
                            not self.input.selected:
                        self.on_select(0, -1)
                    elif event.value == JOYHAT_UP:
                        self.on_select(-1, 0)
                    elif event.value == JOYHAT_RIGHT and\
                            not self.input.selected:
                        self.on_select(0, 1)
                    elif event.value == JOYHAT_DOWN:
                        self.on_select(1, 0)
                elif event.type == pygame.JOYBUTTONDOWN\
                        and event.button == 0 and self.layout.selection:
                    # Select button pressed
                    self.on_key_down(self.layout.selection)

    def on_select(self, increment_row, increment_col, key=None):
        """"Change the currently selected key.

        Parameters
        ----------
        increment_row:
            Of how many rows to move (-1 to go up, 1 to go down).
        increment_col:
            Of how many columns to move (-1 to go left, 1 to go right).
        key:
            The key to start from, by default the currently selected one.
        """
        if self.joystick_navigation:
            if not self.layout.selection:
                if key:
                    self.layout.selection = key
                elif increment_row > 0:
                    self.layout.selection = self.layout.rows[0].keys[0]
                else:
                    self.layout.selection = self.layout.rows[-1].keys[0]
                self.layout.selection.set_selected(1)
                self.input.set_selected(0)
                return

            closest = self.layout.get_key_closest(
                key or self.layout.selection,
                loop_col=not self.input.is_enabled())

            self.layout.selection.set_selected(0)
            self.layout.selection = None
            if closest[(increment_row, increment_col)]:
                self.layout.selection = closest[(increment_row, increment_col)]
                self.layout.selection.set_selected(1)
            else:
                self.input.set_selected(1)

    def on_event(self, event):
        """Deprecated method, only for backward compatibility."""
        self.update((event,))
        self.draw()

    def on_uppercase(self):
        """Uppercase key press handler."""
        self.uppercase = not self.uppercase
        for layout in self.layouts:
            layout.set_uppercase(self.uppercase)

    def on_special_char(self):
        """Special char key press handler."""
        if len(self.layouts) > 1:
            self.special_char = not self.special_char
            if self.special_char:
                self.set_layout(self.layouts[1])
            else:
                self.set_layout(self.layouts[0])

    def on_key_down(self, key):
        """Process key down event by pressing the given key.

        Parameters
        ----------
        key:
            Key that receives the key down event.
        """
        if isinstance(key, vkeys.VBackKey):
            self.input.delete_at_cursor()
        else:
            text = key.update_buffer('')
            if text:
                self.input.add_at_cursor(text)

        if not isinstance(key, vkeys.VActionKey):
            self.text_consumer(self.input.text)



#!/usr/bin/env python
# coding: utf8

"""Simple keyboard usage using custom numeric layout."""

# pylint: disable=import-error
import pygame
import pygame_vkeyboard as vkboard
# pylint: enable=import-error


def on_key_event(text):
    """ Print the current text. """
    print('Current text:', text)


def main(test=False):
    """ Main program.

    :param test: Indicate function is being tested
    :type test: bool
    :return: None
    """

    # Init pygame
    pygame.init()
    screen = pygame.display.set_mode((400, 400))

    # Create keyboard
    model = ['123', '456', '789', '*0#']
    layout = vkboard.VKeyboardLayout(model,
                                     key_size=30,
                                     padding = 15,
                                     height_ratio=0.8,
                                     allow_uppercase=False,
                                     allow_special_chars=False,
                                     allow_space=False)
    keyboard = vkboard.VKeyboard(screen, on_key_event, layout,
                                 joystick_navigation=True)

    # Main loop
    while True:

        events = pygame.event.get()

        for event in events:
            if event.type == pygame.QUIT:
                exit()

        keyboard.update(events)
        rects = keyboard.draw(screen)

        # Flip only the updated area
        pygame.display.update(rects)

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding: utf8

"""Simple keyboard usage using AZERTY layout."""

# pylint: disable=import-error
import pygame
import pygame_vkeyboard as vkboard
# pylint: enable=import-error


def on_key_event(text):
    """ Print the current text. """
    print('Current text:', text)


def main(test=False):
    """ Main program.

    :param test: Indicate function is being tested
    :type test: bool
    :return: None
    """

    # Init pygame
    pygame.init()
    screen = pygame.display.set_mode((500, 400))

    # Create keyboard
    layout = vkboard.VKeyboardLayout(vkboard.VKeyboardLayout.AZERTY, height_ratio=1)
    keyboard = vkboard.VKeyboard(screen, on_key_event, layout)

    clock = pygame.time.Clock()

    # Main loop
    while True:
        clock.tick(100)  # Ensure not exceed 100 FPS

        for event in pygame.event.get():
            keyboard.on_event(event)
            if event.type == pygame.QUIT:
                print("Average FPS: ", clock.get_fps())
                exit()

        # Flip the entire surface
        pygame.display.flip()

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding: utf8

""" Distribution script. """

import sys
from os.path import abspath, dirname, join
from setuptools import setup, find_packages
if sys.version_info[0] == 2:
    from io import open  # Support for encoding on Python 2.x

import pygame_vkeyboard

here = dirname(abspath(__file__))
with open(join(here, 'README.md'), 'r', encoding='utf-8') as stream:
    readme = stream.read()

setup(
    name='pygame-vkeyboard',
    version=pygame_vkeyboard.__version__,
    description=pygame_vkeyboard.__doc__,
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Felix Voituret, Antoine Rousseaux',
    author_email='felix@voituret.fr, anxuae-prog@yahoo.fr',
    url='https://github.com/Faylixe/pygame-vkeyboard',
    download_url='https://pypi.org/project/pygame-vkeyboard/#files',
    license='Apache License 2.0',
    keywords=['pygame', 'keyboard'],
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: pygame'
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=2.7',
    install_requires=[
        'pygame',
    ],
    setup_requires=[
        'setuptools>=41.0.1',
        'wheel>=0.33.4'
    ],
    options={
        'bdist_wheel':
            {'universal': True}  # Support Python 2.x
    },
)



#!/usr/bin/env python
# coding: utf8

"""Simple keyboard usage using QWERTY layout and input text."""

# pylint: disable=import-error
import pygame
import pygame_vkeyboard as vkboard
# pylint: enable=import-error


def on_key_event(text):
    """ Print the current text. """
    print('Current text:', text)


def main(test=False):
    """ Main program.

    :param test: Indicate function is being tested
    :type test: bool
    :return: None
    """

    # Init pygame
    pygame.init()
    screen = pygame.display.set_mode((500, 300), pygame.RESIZABLE)
    screen.fill((100, 100, 100))

    # Create keyboard
    layout = vkboard.VKeyboardLayout(vkboard.VKeyboardLayout.QWERTY,
                                     allow_special_chars=False,
                                     allow_space=False)
    keyboard = vkboard.VKeyboard(screen,
                                 on_key_event,
                                 layout,
                                 renderer=vkboard.VKeyboardRenderer.DARK,
                                 show_text=True,
                                 joystick_navigation=True)

    clock = pygame.time.Clock()

    # Main loop
    while True:
        clock.tick(100)  # Ensure not exceed 100 FPS

        events = pygame.event.get()

        for event in events:
            if event.type == pygame.QUIT:
                print("Average FPS: ", clock.get_fps())
                exit()
            if event.type == pygame.VIDEORESIZE:
                screen.fill((100, 100, 100))

        keyboard.update(events)
        rects = keyboard.draw(screen)

        # Flip only the updated area
        pygame.display.update(rects)

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()
